// $Log$
// Revision 1.10  2006/11/17 00:17:29  stuart
// Compile on CentOS-4
//
// Revision 1.9  2004/02/09 17:28:53  stuart
// Port from libg++ to STL
//
// Revision 1.8  2003/03/24 21:42:35  stuart
// Fix error routing bug cause by change in EDIsink api for edimsg.cc 1.5
//
// Revision 1.7  2001/02/20 22:08:57  stuart
// improve debugging
//
// Revision 1.6  1997/03/26  17:32:38  stuart
// Partners code external
// Check PARTNERS.NOACK
//
// Revision 1.5  1994/08/27  00:08:25  stuart
// check versions and reject when not found
//
// Revision 1.4  1994/07/28  19:15:28  stuart
// show usage before opening files
// replace Array<String> with Segment
//
// Revision 1.3  1993/12/07  19:58:56  stuart
// a few diagnostics
//
// Revision 1.2  1993/11/17  16:25:10  stuart
// usage, multiple receivers, fix FA reception handling
//
#include <fstream>
#include <iostream>
#include <stdlib.h>
#include "edi.h"
#include "edifile.h"
#include "partner.h"

// a generic control class that watches for a begin/end pair
class EDIctl: public EDIsink {
  const char *begin, *end;
protected:
  EDIsink *next;
public:
  EDIctl(const char *b, const char *e) { begin = b; end = e; next = 0; }
  virtual void open(Segment &) = 0;
  virtual void finish() = 0;
  virtual void close(Segment &) = 0;
  int write(Segment &);
};

// the ISA/IEA control class
class ISAhdr: public EDIctl {
  ISAhdr(const ISAhdr &);
  ISAhdr&operator=(const ISAhdr &);
  ISA hdr;
  string routeid;
  Segment &us;
  btFile hdrf;
public:
  ISAhdr(const string &s,Segment &);
  void open(Segment &fld);
  void finish();
  void close(Segment &fld);
  ~ISAhdr() { finish(); }
};

// a convenient way to append error codes to an AKx record
class AKhdr: public EDIctl {
protected:
  AKhdr(const char *b,const char *e): EDIctl(b,e) { }
  Segment err;	// the error/ack record for each group
  void warn(char);	// report an error for the current group
  void reject(char);	// reject the current group (fatal error)
};

// create an SThdr or EDIfile for each GS/GE pair
// create an STfile for Functional Acknowledgment for each SThdr
class GShdr: public AKhdr {
  GShdr(const GShdr &);
  GShdr&operator=(const GShdr &);
  GS grp;
  btFile grpf;
  Segment &us;
  long accepted;
  char passthrough;
public:
  GShdr(Segment &);
  void open(Segment &fld);
  void error(Segment &fld);
  void finish();
  void close(Segment &fld);
  ~GShdr() { finish(); }
};

// create an EDIfile for each ST/SE pair
class SThdr: public AKhdr {
  string prefix, ctlno;
  EDIfile *file;
  SegDict *version;
public:
  SThdr(const string &s,SegDict *vers);
  void open(Segment &fld);
  void finish();
  void close(Segment &fld);
  ~SThdr() { finish(); }
};

PartnerMap pmap;

SThdr::SThdr(const string &s,SegDict *vers):
  AKhdr("ST","SE"), version(vers)
{
  prefix = s; file = 0;
}

int EDIctl::write(Segment &fld) {
  if (fld[0] == begin) {
    finish();
    ++cnt;
    cerr << begin << ": ";
    open(fld);
    return 0;
  }
  if (fld[0] == end) {
    close(fld);
    cerr << end << endl;
    return 0;
  }
  if (next)
    return next->write(fld);
  cerr << begin << ": unexpected: " << fld[0] << endl;
  return -1;	// FIXME: how to report this error
}

ISAhdr::ISAhdr(const string &s,Segment &n):
  EDIctl("ISA","IEA"), routeid(s), us(n),
  hdrf("MESSAGE",BTRDWR,Keydata(&hdr,sizeof hdr))
{ }

void ISAhdr::open(Segment &fld) {
  // do something with Interchange fields
  hdr.authqual = fld[1];
  hdr.authinfo = fld[2];
  hdr.secqual  = fld[3];
  hdr.secinfo  = fld[4];
  hdr.sendqual = fld[5];
  hdr.sender   = fld[6];
  hdr.recvqual = fld[7];
  hdr.receiver = fld[8];
  hdr.time	 = toDateTime(fld[9],fld[10]);
  if (fld[11].length())
    hdr.standard = fld[11][0];
  hdr.version  = fld[12];
  hdr.ctlno	 = atol(fld[13].c_str());
  if (fld[14].length())
    hdr.ackreq	 = fld[14][0];
  if (fld[15].length())
    hdr.test	 = fld[15][0];
  hdr.grps	 = 0;
  string dest(hdr.recvqual+"-"+string(hdr.receiver));
  cerr	<< '\'' << hdr.sendqual+"-"+string(hdr.sender) << '\''
	<< " --> '" << dest << '\''
	<< endl;
  if (dest == routeid)
    next = new GShdr(us);
  else {	// route to new destination
    cerr << dest << " != " << routeid << endl;
    cerr << "ISA: routing not implemented\n";
#if 0
    if (partidf.readeq(&hdr.sendqual + &hdr.sender))
      ; // generate ISA routing error
    string nodename = partrec.id;
    os = procstream("smail "+nodename+"!edi");
    next = EDIstream(*os,config.elemsep,config.segsep,config.subsep);
#endif
  }
}

void ISAhdr::close(Segment &fld) {
  if (!next) return;
  hdr.grps = next->cnt;
  delete next; next = 0;
  if (hdr.grps != atol(fld[1].c_str()))
    cerr << "ISA: group count does not match\n";
  if (hdr.ctlno != atol(fld[2].c_str()))
    cerr << "ISA: control number does not match\n";
  hdrf.write();
  if (hdr.ackreq == 'Y')
    return;	// generate ISA ack
}

void ISAhdr::finish() {
  if (next) {
    cerr << "ISA: missing IEA segment\n";
    // generate ISA error
    Segment seg;
    close(seg +"IEA"+dec(next->cnt)+dec(long(hdr.ctlno)));
  }
}

GShdr::GShdr(Segment &s):
  AKhdr("GS","GE"), us(s),
  grpf("GROUP",BTRDWR,Keydata(&grp,sizeof grp))
{ }

void AKhdr::warn(char c) {
  if (err[1] == "A")
    err[1] = 'E';
  err.add(string(1,c));
}

void AKhdr::reject(char c) {
  err[1] = 'R';
  err.add(string(1,c));
}

void GShdr::finish() {
  if (next) {
    warn('3');	// missing trailer
    // supply missing trailer
    Segment seg;
    close(seg+"GE"+dec(next->cnt)+dec(long(grp.ctlno)));
  }
}

void GShdr::open(Segment &fld) {
  // check whether version is supported
  SegDict *segvers = SegDict::library.findVersion(fld[8]);
  grp.func = fld[1];
  grp.sndr = fld[2];
  grp.rcvr = fld[3];
  grp.time = toDateTime(fld[4],fld[5]);
  grp.ctlno = atol(fld[6].c_str());
  grp.agency = fld[7];
  grp.version = fld[8];
  grp.sets = 0;		// transaction sets received
  accepted = 0;		// transaction sets accepted
  errs = 0;
  string fromcode(pmap[fld[2]]),tocode(pmap[fld[3]]);
  string spname(tocode+'/'+fromcode+'.'+fld[1]);
  cerr << spname << endl;
  for (int i = 0; i < us.size(); ++i) {
    if (tocode == us[i]) {
      passthrough = 0;
      // SThdr creates application spool files
      next = new SThdr(spname,segvers);
      // create response file for reporting acks/errs
      if (fld[1] != "FA") {	// don't xmit ACK to ACK
	static FCHAR<4> id;
	static FCHAR<2> msg;
	static btFile noackf(&id + &msg);
	if (noackf.keylen() == 0)
	  noackf.open("PARTNERS.NOACK",BTRDONLY+NOKEY);
	id = fromcode;
	if (noackf.readeq(&id + &grp.func)) {
	  // dropback to specific version for reject when we don't recognize it
	  SegDict *errvers = segvers ? segvers
				   : SegDict::library.findVersion("003020");
	  if (errvers)
	    errs = new STfile("997",fromcode+'/'+tocode+".FA",errvers);
	}
      }
      Segment seg;
      error(seg+"AK1"+fld[1]+fld[6]);
      next->errors(errs);
      // initialize group err/ack record
      err.clear()+"AK9"+'A'+'0'+'0'+'0';
      if (!segvers)
	reject('2');		// unsupported version
  //  if (trangrpf.readeq(&grp.func))
  //    reject('1');		// unsupported function group
      return;
    }
  }
  // if not for us, spool group verbatim
  passthrough = 1;
  next = new EDIfile(spname + fld[6],segvers);
  next->write(fld);	// include GS/GE segments
}

void GShdr::error(Segment &fld) {
  if (fld[0] == "AK5") {
    if (fld[1] != "R")
      ++accepted;
  }
  AKhdr::error(fld);
}

void GShdr::close(Segment &fld) {
  if (!next) {	// GE with no GS
    return;	// ignore
  }
  grp.sets = next->cnt;
  if (passthrough) {
    next->write(fld);	// GE record to spool file
    delete next;	// complete spooled transactions
  }
  else {
    long sent = atol(fld[1].c_str());
    long received = next->cnt;
    delete next;	// complete spooled transactions
    if (received != accepted && err[1] == "A")
      err[1] = 'P';
    if (err[1] == "R")
      accepted = 0;
    err[2] = fld[1];	// sent
    err[3] = dec(received);
    err[4] = dec(accepted);
    if (received != sent)
      warn('5');	// transaction set count does not match
    if (atol(fld[2].c_str()) != grp.ctlno)
      reject('4');	// control number does not match
    error(err);
    grpf.write();
  }
  next = 0;
  delete errs;	// complete response transaction
  errs = 0;
}

void SThdr::open(Segment &fld) {
  ctlno = "";
  if (fld.size() > 2)
    ctlno = fld[2];
  if (ctlno.empty())
    reject('7');	// missing control no.
  cerr << prefix + '.' + ctlno << endl;
  file = new EDIfile(prefix+'.'+ctlno,version);
  file->errors(errs);
  next = new SegmentChecker(file);
  next->errors(errs);
  error(err.clear()+"AK2"+fld[1]+ctlno);
  err.clear()+"AK5"+'A';
  if (next->write(fld))
    reject('6');	// Checker won't accept transet id
}

void SThdr::finish() {
  if (next) {
    warn('2');	// missing trailer
    // supply missing trailer
    Segment seg;
    close(seg+"SE"+dec(next->cnt + 1)+ctlno);
  }
}

void SThdr::close(Segment &fld) {
  if (!next)
    return;	// SE with no ST
  next->write(fld);
  if (atol(fld[1].c_str()) != file->cnt)
    warn('4');		// segment count does not match
  if (fld[2] != ctlno)
    reject('3');	// transaction set control number does not match
  error(err);
  delete next;
  next = 0;
  delete file;
  file = 0;
}

int main(int argc,char **argv) {
  if (argc < 3) {
    cerr << "\t$Id$\n\
Usage:	edirecv [-v] routing-dest id [id ...] <message\n\
	-v		show more debugging info\n\
	routing-dest	ISA qual-destination, e.g. 01-003423001\n\
	id		partner ids to process messages for\n\
			groups for other ids are passed through unchanged\n\
\n\
Files:	BTASDIR				BTAS Spool Directory\n\
	/bms/edi/*.seg			UNIX Segment Tables for versions\n\
	/edx/AIRCO/EDI/CODE.MAST	BTAS Code Dictionary\n\
	MESSAGE				BTAS ISA header log\n\
	GROUP				BTAS GS header log\n\
	PARTNERS			BTAS EDI addresses\n\
	PARTNERS.NOACK			BTAS suppress FA to these partners\n\
";
    return 1;
  }
  const char *interid = argv[1];
  Segment fld, rcvr;
  for (int i = 2; i < argc; ++i)
    rcvr.add(argv[i]);
  EdiMsg msg(cin);
  while (msg.getISA("ISA",fld)) {
    cerr << "ElemSep = " << hex << (int)msg.sep
	 << ", SegSep = " << hex << (int)msg.segsep
	 << ", SubSep = " << hex << (int)msg.subsep
	 << endl;
    ISAhdr hdr(interid,rcvr);
    // see if message is for this machine
    do
      hdr.write(fld);
    while (fld[0] != "IEA" && msg.getseg(fld));
  }
  return 0;
}
